/* Catweasel test program
   From cwfloppy-0.2.1 package, copyright (c) 1998 Michael Krause
   Modified by Timothy Mann
   $Id: testhist.c,v 1.5 2001/07/14 22:48:13 mann Exp $

   Reads a track and prints a histogram of the samples (flux
   transition interarrival times) returned.  Also categorizes them
   (assuming MFM encoding, and with fixed thresholds) and prints the
   mean and standard deviation of the actual length of samples in each
   category.  Also optionally dumps the samples to a file.

   The track is read NNN times (defined below).  When samples are
   being dumped to a file, each read is preceded by a pseudo-sample
   equal to 0x80 + read number (0-origin).  The actual samples can be
   at most 0x7f.

   Usage: testhist port drive track side clock [file]

   Clock is normally 0 to read SD/DD, 1 to read HD; this setting makes
   the category thresholds reasonable.  
*/

#define NNN 4
#define CWHZ 14161000.0

#undef __KERNEL__
#define __LOWLEVEL__
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#if linux
#include <sys/io.h>
#endif

#if __DJGPP__
#include <pc.h>
#include <dpmi.h>
#define inb(p) inportb(p)
#define outb(v, p) outportb(p, v) /* args in opposite order, arrgh */
#endif

#if DEBUG
unsigned char INB(unsigned short port)
{
  unsigned char b = inb(port);
  printf("      %04x => %02x\n", port, b);
  return b;
}
void OUTB(unsigned char b, unsigned short port)
{
  printf("%02x => %04x\n", b, port);
  outb(b, port);
}
#define inb INB
#define outb OUTB
#endif /*DEBUG*/

#include "cwfloppy.h"
#include "catweasl.c"

FILE *f;
struct catweasel_contr c;
unsigned char trackbuffer[18*512];

#if __DJGPP__
/* DJGPP's usleep is based on a clock with a 55 ms granularity, so
   it's not usable for short sleeps!  Fortunately DJGPP also makes
   a fine-grained clock available.  This clock works only on
   MS-DOS (with a DPMI host) or Windows 9X, not Windows 3.1 or NT.
   Of course cw2dmk won't work on NT anyway, because it accesses
   I/O ports directly.
*/
#include <time.h>
unsigned int
usleep(unsigned int _useconds)
{
  uclock_t wakeup = uclock() +
    ((uclock_t)_useconds) * ((uclock_t)UCLOCKS_PER_SEC) / 1000000LL;
  while (uclock() <= wakeup) __dpmi_yield();
  return 0;
}
#endif

void ms_delay(int x) {
    usleep(1000*x);
}

int cwclock = 1;
int readtime = -1;

static void histo_track(int drive, int track, int side, unsigned int *buf) {
    char b;
    int i;

    catweasel_seek(&c.drives[drive], track);

    memset(buf, 0, 128*sizeof(int));
    for(i=0;i<NNN;i++) {
	catweasel_read(&c.drives[drive], side, cwclock, readtime, 0);
	if (f) putc(0x80 + i, f);
	while((b = inb(c.iobase)) >= 0) {
	    buf[(int)b]++;
	    if (f) putc(b, f);
	}
    }
}

static void eval_histo(unsigned int *histogram)
{
    int i, ii, j, pwidth, psamps, psampsw;
    double peak[3], sd[3], ps[3];

    /* Print histogram */
    for(i=0;i<128;i+=8) {
	printf("%3d: %06d %06d %06d %06d %06d %06d %06d %06d\n", i,
	       histogram[i+0], histogram[i+1], histogram[i+2], histogram[i+3],
	       histogram[i+4], histogram[i+5], histogram[i+6], histogram[i+7]);
    }

    /* Find two or three peaks */
    i = 0;
    for (j=0; j<3; j++) {
      pwidth = 0;
      psamps = 0;
      psampsw = 0;
      while (histogram[i] < 64 && i < 128) i++;
      while (histogram[i] >= 64 && i < 128) {
	pwidth++;
	psamps += histogram[i];
	psampsw += histogram[i] * i;
	i++;
      }
      if (pwidth == 0 || pwidth > 16) {
	/* Not a real peak */
	break;
      } else {
	peak[j] = ((double) psampsw) / psamps; 
	ps[j] = psamps;
	sd[j] = 0.0;
	for (ii = i - pwidth; ii < i; ii++) {
	  sd[j] += histogram[ii] * pow((double)ii - peak[j], 2);
	}
	sd[j] = sqrt(sd[j]/((double)psamps-1));
	printf("peak %d: mean %.05f, sd %.05f\n", j, peak[j], sd[j]);
      }
    }

    /* Guess drive RPM based on total number of cw clock cycles */
    /* Note that the sample values and histogram bucket numbers
       are (I believe) 1 less than the actual number of cycles, so
       we need to add 1 in some places here */
    psamps = 0;
    psampsw = 0;
    for (i=0; i<128; i++) {
      psamps += histogram[i];
      psampsw += histogram[i] * (i + 1);
    }
    printf("drive speed approx    %f RPM\n",
	   CWHZ / (cwclock + 1) / psampsw * NNN * 60.0);
    
    /* Guess bit clock by imputing the expected number of
       clocks to each peak */
#if 0
    /* This code weights each peak the same */
    if (j == 2) {
      /* FM encoding */
      printf("FM data clock approx  %f kHz\n",
	     CWHZ / 1000.0 /  (cwclock + 1) /
	     ((peak[0]/0.5 + peak[1]/1.0)/2.0 + 1.0));
    } else if (j == 3) {
      /* MFM encoding */
      printf("MFM data clock approx %f kHz\n",
	     CWHZ / 1000.0 /  (cwclock + 1) /
	     ((peak[0]/1.0 + peak[1]/1.5 + peak[2]/2.0)/3.0 + 1.0));
    }      
#else
    /* This code weights each peak by the number of samples in it;
       I think that should be a bit more accurate. */
    if (j == 2) {
      /* FM encoding */
      printf("FM data clock approx  %f kHz\n",
	     CWHZ / 1000.0 / (cwclock + 1) /
	     ((peak[0]/0.5*ps[0] + peak[1]/1.0*ps[1])
	      / (ps[0] + ps[1]) + 1.0));
    } else if (j == 3) {
      /* MFM encoding */
      printf("MFM data clock approx %f kHz\n",
	     CWHZ / 1000.0 / (cwclock + 1) /
	     ((peak[0]/1.0*ps[0] + peak[1]/1.5*ps[1] + peak[2]/2.0*ps[2])
	      / (ps[0] + ps[1] + ps[2]) + 1.0));
    }      
#endif
}


int main(int argc, char **argv) {
    int track, side, port, drive;
    unsigned int buf[128];

    if(argc != 6 && argc != 7) {
	fprintf(stderr,
		"Usage: %s port drive track side clock [file]\n", argv[0]);
	return 1;
    }

    port = strtol(argv[1], NULL, 0);

#if linux
    if(ioperm(port, 8, 1) == -1) {
	fprintf(stderr, "No access to IO ports\n");
	return 1;
    }
    setuid(getuid());
#endif

    drive = atoi(argv[2]);
    track = atoi(argv[3]);
    side = atoi(argv[4]);
    cwclock = atoi(argv[5]);
    if (argc == 7) {
      if (strcmp(argv[6], "-") == 0) {
	f = stdout;
      } else {
	f = fopen(argv[6], "w");
      }
    }

    c.iobase = port;
    c.msdelay = ms_delay;
    catweasel_init_controller(&c);

    catweasel_select(&c, !drive, drive);
    catweasel_set_motor(&c.drives[drive], 1);
    ms_delay(500);

    printf("Reading track %d, side %d...\n", track, side);
    histo_track(drive, track, side, buf);
    eval_histo(buf);

    catweasel_set_motor(&c.drives[drive], 0);
    catweasel_select(&c, 0, 0);
    
    catweasel_free_controller(&c);

    if (f) fclose(f);
    
    return 0;
}
